//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Device/Data/DataUtils.h
//! @brief     Defines namespace DataUtils.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_DEVICE_DATA_DATAUTILS_H
#define BORNAGAIN_DEVICE_DATA_DATAUTILS_H

#include "Device/Data/Powerfield.h"
#include <memory>

namespace DataUtils::Data {

//! Returns relative difference between two data sets sum(dat[i] - ref[i])/ref[i]).
double relativeDataDifference(const std::vector<double>& dat, const std::vector<double>& ref);

//! Transforms coordinate on axis into the bin-fraction-coordinate.
double coordinateToBinf(double coordinate, const IAxis& axis);

//! Transforms bin-fraction-coordinate into axis coordinate.
double coordinateFromBinf(double value, const IAxis& axis);

//! Transforms x,y coordinate from Powerfield axes coordinates to bin-fraction-coordinates.
void coordinateToBinf(double& x, double& y, const Powerfield<double>& data);

//! Transforms x,y coordinate from bin-fraction-coordinates to Powerfield's axes coordinates.
void coordinateFromBinf(double& x, double& y, const Powerfield<double>& data);

//! Creates a vector of vectors of double (2D Array) from Powerfield.
std::vector<std::vector<double>> create2DArrayfromPowerfield(const Powerfield<double>& data);

//! Reads 1D array of doubles to Python, for use in persistence test
Powerfield<double>* importArrayToPowerfield(const std::vector<double>& vec);
//! Reads 2D array of doubles to Python, for use in persistence test
Powerfield<double>* importArrayToPowerfield(const std::vector<std::vector<double>>& vec);

#ifndef SWIG

std::unique_ptr<Powerfield<double>>
createRelativeDifferenceData(const Powerfield<double>& data, const Powerfield<double>& reference);

//! Returns new object with input data rotated by
//! n*90 deg counterclockwise (n > 0) or clockwise (n < 0)
//! Axes are swapped if the data is effectively rotated by 90 or 270 degrees
//! Applicable to 2D arrays only
std::unique_ptr<Powerfield<double>> createRearrangedDataSet(const Powerfield<double>& data, int n);

//! Creates Powerfield from a 2D Array.
std::unique_ptr<Powerfield<double>>
createPowerfieldfrom2DArray(const std::vector<std::vector<double>>& array_2d);

//! Creates Fourier Transform (Powerfield format) of intensity map (Powerfield format).
std::unique_ptr<Powerfield<double>> createFFT(const Powerfield<double>& data);

#endif // SWIG

} // namespace DataUtils::Data

#endif // BORNAGAIN_DEVICE_DATA_DATAUTILS_H
