//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Device/Data/LLData.h
//! @brief     Defines class LLData.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef USER_API
#ifndef BORNAGAIN_DEVICE_DATA_LLDATA_H
#define BORNAGAIN_DEVICE_DATA_LLDATA_H

#include <algorithm>
#include <limits>
#include <numeric>
#include <stdexcept>

//! Template class to store data of any type in multi-dimensional space (low-level).

template <class T>
class LLData {
public:
    // construction, destruction and assignment
    LLData(size_t rank, const int* sizes);
    LLData(const LLData<T>& right);
    LLData<T>& operator=(const LLData<T>& right);
    ~LLData();

    LLData<double> meanValues() const;

    // accessors
    T& operator[](size_t i);
    const T& operator[](size_t i) const;

    // arithmetic operations
    LLData<T>& operator+=(const LLData<T>& right);
    LLData<T>& operator-=(const LLData<T>& right);
    LLData<T>& operator*=(const LLData<T>& right);
    LLData<T>& operator/=(const LLData<T>& right);

    // initialization, scaling
    void setAll(const T& value);

    // retrieve basic info
    size_t getTotalSize() const;
    inline size_t rank() const { return m_rank; }
    const int* sizes() const { return m_sizes; }
    T getTotalSum() const;
    T max() const;

private:
    void allocate(size_t rank, const int* sizes);
    void clear();
    bool checkSizes(size_t rank, const int* sizes) const;
    void swapContents(LLData<T>& other);
    T getZeroElement() const;

    size_t m_rank;
    int* m_sizes;
    T* m_data_array;
};

#ifndef SWIG

// Global helper functions for arithmetic
template <class T>
LLData<T> operator+(const LLData<T>& left, const LLData<T>& right);
template <class T>
LLData<T> operator-(const LLData<T>& left, const LLData<T>& right);
template <class T>
LLData<T> operator*(const LLData<T>& left, const LLData<T>& right);
template <class T>
LLData<T> operator/(const LLData<T>& left, const LLData<T>& right);

// Global helper functions for comparison
template <class T>
bool HaveSameSizes(const LLData<T>& left, const LLData<T>& right);

template <class T>
inline LLData<T>::LLData(size_t rank, const int* sizes)
    : m_rank(0)
    , m_sizes(nullptr)
    , m_data_array(0)
{
    allocate(rank, sizes);
}

template <class T>
LLData<T>::LLData(const LLData<T>& right)
    : m_rank(0)
    , m_sizes(nullptr)
    , m_data_array(0)
{
    allocate(right.rank(), right.sizes());
    for (size_t i = 0; i < getTotalSize(); ++i)
        m_data_array[i] = right[i];
}

template <class T>
LLData<T>::~LLData()
{
    clear();
}

template <class T>
LLData<T>& LLData<T>::operator=(const LLData<T>& right)
{
    if (this != &right) {
        LLData<T> copy(right);
        swapContents(copy);
    }
    return *this;
}

template <class T>
inline T& LLData<T>::operator[](size_t i)
{
    return m_data_array[i];
}

template <class T>
inline const T& LLData<T>::operator[](size_t i) const
{
    return m_data_array[i];
}

template <class T>
LLData<T>& LLData<T>::operator+=(const LLData<T>& right)
{
    if (!HaveSameSizes(*this, right))
        throw std::runtime_error(
            "Operation += on LLData requires both operands to have the same sizes");
    for (size_t i = 0; i < getTotalSize(); ++i)
        m_data_array[i] += right[i];
    return *this;
}

template <class T>
LLData<T>& LLData<T>::operator-=(const LLData& right)
{
    if (!HaveSameSizes(*this, right))
        throw std::runtime_error(
            "Operation -= on LLData requires both operands to have the same sizes");
    for (size_t i = 0; i < getTotalSize(); ++i)
        m_data_array[i] -= right[i];
    return *this;
}

template <class T>
LLData<T>& LLData<T>::operator*=(const LLData& right)
{
    if (!HaveSameSizes(*this, right))
        throw std::runtime_error(
            "Operation *= on LLData requires both operands to have the same sizes");
    for (size_t i = 0; i < getTotalSize(); ++i)
        m_data_array[i] *= right[i];
    return *this;
}

template <class T>
LLData<T>& LLData<T>::operator/=(const LLData& right)
{
    if (!HaveSameSizes(*this, right))
        throw std::runtime_error(
            "Operation /= on LLData requires both operands to have the same sizes");
    for (size_t i = 0; i < getTotalSize(); ++i) {
        double ratio;
        if (std::abs(m_data_array[i] - right[i])
            <= std::numeric_limits<double>::epsilon() * std::abs(right[i])) {
            ratio = 1.0;
        } else if (std::abs(right[i]) <= std::numeric_limits<double>::min())
            ratio = double(m_data_array[i]) / std::numeric_limits<double>::min();
        else
            ratio = double(m_data_array[i] / right[i]);
        m_data_array[i] = (T)ratio;
    }
    return *this;
}

template <class T>
void LLData<T>::setAll(const T& value)
{
    std::fill(m_data_array, m_data_array + getTotalSize(), value);
}

template <class T>
inline size_t LLData<T>::getTotalSize() const
{
    int result = std::accumulate(m_sizes, m_sizes + m_rank, 1, std::multiplies<int>{});
    return static_cast<size_t>(result);
}

template <class T>
T LLData<T>::getTotalSum() const
{
    return std::accumulate(m_data_array, m_data_array + getTotalSize(), getZeroElement());
}

template <class T>
T LLData<T>::max() const
{
    return *std::max_element(m_data_array, m_data_array + getTotalSize());
}

template <class T>
void LLData<T>::allocate(size_t rank, const int* sizes)
{
    clear();
    if (!checkSizes(rank, sizes))
        throw std::runtime_error("LLData<T>::allocate error: sizes must be > 0");

    m_rank = rank;
    if (m_rank) {
        m_sizes = new int[m_rank];
        std::copy(sizes, sizes + rank, m_sizes);
        m_data_array = new T[getTotalSize()];
    } else {
        m_data_array = new T[1];
    }
}

template <class T>
void LLData<T>::clear()
{
    if (m_rank > 0) {
        m_rank = 0;
        delete[] m_data_array;
        delete[] m_sizes;
        m_data_array = nullptr;
        m_sizes = nullptr;
    } else {
        delete[] m_data_array;
    }
}

template <class T>
inline bool LLData<T>::checkSizes(size_t rank, const int* sizes) const
{
    return std::all_of(sizes, sizes + rank, [](const int& dim) -> bool { return dim > 0; });
}

template <class T>
void LLData<T>::swapContents(LLData<T>& other)
{
    std::swap(this->m_rank, other.m_rank);
    std::swap(this->m_sizes, other.m_sizes);
    std::swap(this->m_data_array, other.m_data_array);
}

template <class T>
T LLData<T>::getZeroElement() const
{
    T result = 0;
    return result;
}

template <class T>
LLData<T> operator+(const LLData<T>& left, const LLData<T>& right)
{
    auto* p_result = new LLData<T>(left);
    (*p_result) += right;
    return *p_result;
}

template <class T>
LLData<T> operator-(const LLData<T>& left, const LLData<T>& right)
{
    auto* p_result = new LLData<T>(left);
    (*p_result) -= right;
    return *p_result;
}

template <class T>
LLData<T> operator*(const LLData<T>& left, const LLData<T>& right)
{
    auto* p_result = new LLData<T>(left);
    (*p_result) *= right;
    return *p_result;
}

template <class T>
LLData<T> operator/(const LLData<T>& left, const LLData<T>& right)
{
    auto* p_result = new LLData<T>(left);
    *p_result /= right;
    return *p_result;
}

template <class T>
bool HaveSameSizes(const LLData<T>& left, const LLData<T>& right)
{
    if (left.rank() != right.rank())
        return false;
    const int* ldims = left.sizes();
    const int* rdims = right.sizes();
    for (size_t i = 0; i < left.rank(); ++i) {
        if (ldims[i] != rdims[i])
            return false;
    }
    return true;
}

#endif // SWIG

#endif // BORNAGAIN_DEVICE_DATA_LLDATA_H
#endif // USER_API
