//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Device/Histo/Histogram1D.cpp
//! @brief     Implements class Histogram1D.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Device/Histo/Histogram1D.h"
#include "Base/Axis/FixedBinAxis.h"
#include "Base/Axis/VariableBinAxis.h"
#include "Base/Py/PyUtils.h"
#include "Device/Data/Powerfield.h"
#include <memory>

Histogram1D::Histogram1D(const Powerfield<double>& data)
    : IHistogram(data.axis(0))
{
    init_from_data(data);
}

Histogram1D::Histogram1D(Powerfield<CumulativeValue>* data)
    : IHistogram(data)
{
}

Histogram1D* Histogram1D::clone() const
{
    return new Histogram1D(*this);
}

std::vector<double> Histogram1D::binCenters() const
{
    return xAxis().binCenters();
}

std::vector<double> Histogram1D::binValues() const
{
    return IHistogram::getDataVector(IHistogram::DataType::INTEGRAL);
}

std::vector<double> Histogram1D::binErrors() const
{
    return IHistogram::getDataVector(IHistogram::DataType::STANDARD_ERROR);
}

#ifdef BORNAGAIN_PYTHON

PyObject* Histogram1D::binCentersNumpy() const
{
    return BaseUtils::Python::createNumpyArray(binCenters());
}

PyObject* Histogram1D::binValuesNumpy() const
{
    return BaseUtils::Python::createNumpyArray(binValues());
}

PyObject* Histogram1D::binErrorsNumpy() const
{
    return BaseUtils::Python::createNumpyArray(binErrors());
}

#endif // BORNAGAIN_PYTHON

Histogram1D* Histogram1D::crop(double xmin, double xmax)
{
    const std::unique_ptr<IAxis> xaxis{xAxis().clone()};
    xaxis->clip(xmin, xmax);
    auto out = new Powerfield<CumulativeValue>(*xaxis);
    size_t iout = 0;
    for (size_t i = 0; i < m_data->allocatedSize(); ++i) {
        const double x = m_data->projectedCoord(i, 0);
        if (xaxis->contains(x))
            (*out)[iout++] = (*m_data)[i];
    }
    return new Histogram1D(out);
}
