//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Device/Histo/Histogram2D.cpp
//! @brief     Implements class Histogram2D.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Device/Histo/Histogram2D.h"
#include "Base/Axis/VariableBinAxis.h"
#include "Device/Data/Powerfield.h"
#include "Device/Histo/Histogram1D.h"
#include <memory>

Histogram2D::Histogram2D(const Powerfield<double>& data)
    : IHistogram(data.axis(0), data.axis(1))
{
    init_from_data(data);
}

Histogram2D::Histogram2D(Powerfield<CumulativeValue>* data)
    : IHistogram(data)
{
}

Histogram2D* Histogram2D::clone() const
{
    return new Histogram2D(*this);
}

Histogram1D* Histogram2D::projectionX()
{
    return create_projectionX(0, static_cast<int>(xAxis().size()) - 1);
}

Histogram1D* Histogram2D::projectionX(double yvalue)
{
    int ybin_selected = static_cast<int>(yAxis().findClosestIndex(yvalue));
    return create_projectionX(ybin_selected, ybin_selected);
}

Histogram1D* Histogram2D::projectionX(double ylow, double yup)
{
    int ybinlow = static_cast<int>(yAxis().findClosestIndex(ylow));
    int ybinup = static_cast<int>(yAxis().findClosestIndex(yup));
    return create_projectionX(ybinlow, ybinup);
}

Histogram1D* Histogram2D::projectionY()
{
    return create_projectionY(0, static_cast<int>(xAxis().size()) - 1);
}

Histogram1D* Histogram2D::projectionY(double xvalue)
{
    int xbin_selected = static_cast<int>(xAxis().findClosestIndex(xvalue));
    return create_projectionY(xbin_selected, xbin_selected);
}

Histogram1D* Histogram2D::projectionY(double xlow, double xup)
{
    int xbinlow = static_cast<int>(xAxis().findClosestIndex(xlow));
    int xbinup = static_cast<int>(xAxis().findClosestIndex(xup));
    return create_projectionY(xbinlow, xbinup);
}

Histogram2D* Histogram2D::crop(double xmin, double ymin, double xmax, double ymax)
{
    const std::unique_ptr<IAxis> xaxis{xAxis().clone()};
    const std::unique_ptr<IAxis> yaxis{yAxis().clone()};
    xaxis->clip(xmin, xmax);
    yaxis->clip(ymin, ymax);

    auto* out = new Powerfield<CumulativeValue>(*xaxis, *yaxis);
    size_t iout = 0;
    for (size_t i_flat = 0; i_flat < m_data->allocatedSize(); ++i_flat) {
        double x = m_data->projectedCoord(i_flat, 0);
        double y = m_data->projectedCoord(i_flat, 1);
        if (xaxis->contains(x) && yaxis->contains(y))
            (*out)[iout++] = (*m_data)[i_flat];
    }
    return new Histogram2D(out);
}

void Histogram2D::setContent(const std::vector<std::vector<double>>& data)
{
    reset();
    addContent(data);
}

void Histogram2D::addContent(const std::vector<std::vector<double>>& data)
{
    auto shape = DataUtils::Array::getShape(data);
    const size_t nrows = shape.first;
    const size_t ncols = shape.second;

    if (nrows != m_data->axis(1).size() || ncols != m_data->axis(0).size()) {
        std::ostringstream ostr;
        ostr << "Histogram2D::addContent() -> Shape of input array [" << nrows << ", " << ncols
             << "] doesn't mach histogram axes. "
             << "X-axis size: " << m_data->axis(0).size()
             << "Y-axis size: " << m_data->axis(1).size();
        throw std::runtime_error(ostr.str());
    }

    for (size_t row = 0; row < nrows; ++row) {
        for (size_t col = 0; col < ncols; ++col) {
            size_t globalbin = nrows - row - 1 + col * nrows;
            (*m_data)[globalbin].add(data[row][col]);
        }
    }
}

Histogram1D* Histogram2D::create_projectionX(int ybinlow, int ybinup)
{
    Powerfield<double> out(this->xAxis());

    for (size_t index = 0; index < getTotalNumberOfBins(); ++index) {
        int ybin = static_cast<int>(yAxisIndex(index));
        if (ybin >= ybinlow && ybin <= ybinup) {
            double x = xAxisValue(index);
            if (!xAxis().contains(x))
                continue;
            size_t iout = xAxis().findClosestIndex(x);
            out[iout] += binContent(index);
        }
    }

    return new Histogram1D(out);
}

Histogram1D* Histogram2D::create_projectionY(int xbinlow, int xbinup)
{
    Powerfield<double> out(this->yAxis());

    for (size_t index = 0; index < getTotalNumberOfBins(); ++index) {
        int xbin = static_cast<int>(xAxisIndex(index));
        if (xbin >= xbinlow && xbin <= xbinup) {

            // TODO: duplicates code from create_projectionX, move to Powerfield<double>

            double y = yAxisValue(index);
            if (!yAxis().contains(y))
                continue;
            size_t iout = yAxis().findClosestIndex(y);
            out[iout] += binContent(index);
        }
    }

    return new Histogram1D(out);
}
