//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Device/Histo/Histogram2D.h
//! @brief     Defines class Histogram2D.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_DEVICE_HISTO_HISTOGRAM2D_H
#define BORNAGAIN_DEVICE_HISTO_HISTOGRAM2D_H

#include "Device/Data/ArrayUtils.h"
#include "Device/Histo/IHistogram.h"

//! Two dimensional histogram.
//! @ingroup tools

class Histogram2D : public IHistogram {
public:
    //! Constructor for 2D histograms from basic Powerfield object
    Histogram2D(const Powerfield<double>& data);

    Histogram2D(Powerfield<CumulativeValue>* data);

    //! Returns clone of histogram
    Histogram2D* clone() const override;

    //! Returns the number of histogram dimensions
    size_t rank() const override { return 2; }

    //! Project a 2D histogram into 1D histogram along X. The projection is made
    //! from all bins along y-axis.
    Histogram1D* projectionX();

    //! @brief Project a 2D histogram into 1D histogram along X. The projection is made
    //! from the y-bin closest to given ordinate yvalue.
    //! @param yvalue the value on y-axis at which projection is taken
    Histogram1D* projectionX(double yvalue);

    //! @brief Project a 2D histogram into 1D histogram along X. The projection is made from
    //! all y-bins corresponding to ordinate between ylow and yup.
    //! @param ylow lower edje on y-axis
    //! @param yup upper edje on y-axis
    Histogram1D* projectionX(double ylow, double yup);

    //! Project a 2D histogram into 1D histogram along Y. The projection is made
    //! from all bins along x-axis.
    Histogram1D* projectionY();

    //! @brief Project a 2D histogram into 1D histogram along Y. The projection is made
    //! from the x-bin closest to given abscissa xvalue.
    //! @param xvalue the value on x-axis at which projection is taken
    Histogram1D* projectionY(double xvalue);

    //! @brief Project a 2D histogram into 1D histogram along Y. The projection is made from
    //! all x-bins corresponding to abscissa between xlow and xup.
    //! @param xlow lower edje on x-axis
    //! @param xup upper edje on x-axis
    Histogram1D* projectionY(double xlow, double xup);

    //! Creates new histogram by applying rectangular clip.
    Histogram2D* crop(double xmin, double ymin, double xmax, double ymax);

    //! Sets the values in histograms channels from numpy array,
    void setContent(const std::vector<std::vector<double>>& data);

    //! Add to values in histograms channels from numpy array,
    void addContent(const std::vector<std::vector<double>>& data);

private:
    //! Creates projection along X. The projections is made by collecting the data in the range
    //! between [ybinlow, ybinup].
    Histogram1D* create_projectionX(int ybinlow, int ybinup);

    //! Creates projection along Y. The projections is made by collecting the data in the range
    //! between [xbinlow, xbinup].
    Histogram1D* create_projectionY(int xbinlow, int xbinup);
};

#endif // BORNAGAIN_DEVICE_HISTO_HISTOGRAM2D_H
