//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Sample/InterferenceItems.h
//! @brief     Defines InterferenceItems's classes
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2021
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_GUI_MODEL_SAMPLE_INTERFERENCEITEMS_H
#define BORNAGAIN_GUI_MODEL_SAMPLE_INTERFERENCEITEMS_H

#include "GUI/Model/Descriptor/DoubleProperty.h"
#include "GUI/Model/Descriptor/SelectionProperty.h"
#include "GUI/Model/Descriptor/UIntProperty.h"
#include "GUI/Model/Sample/Lattice2DItems.h"
#include "GUI/Model/Sample/ProfileItems.h"
#include <memory>

class IInterference;
class Streamer;

class InterferenceItem {
public:
    virtual ~InterferenceItem() = default;
    virtual std::unique_ptr<IInterference> createInterference() const = 0;
    virtual void serialize(Streamer& s) = 0;

    DoubleDescriptor positionVariance() const;

protected:
    InterferenceItem();

    DoubleProperty m_positionVariance;
};

// ------------------------------------------------------------------------------------------------

class Interference1DLatticeItem : public InterferenceItem {
public:
    Interference1DLatticeItem();

    std::unique_ptr<IInterference> createInterference() const override;
    void serialize(Streamer& s) override;

    DoubleDescriptor length() const;
    DoubleDescriptor rotationAngle() const;

    void setDecayFunction(Profile1DItem* p);
    SelectionDescriptor<Profile1DItem*> decayFunction() const;

private:
    DoubleProperty m_length;
    DoubleProperty m_rotationAngle;
    SelectionProperty<Profile1DItem*> m_decayFunction;
};

// ------------------------------------------------------------------------------------------------

class Interference2DAbstractLatticeItem : public InterferenceItem {
public:
    SelectionDescriptor<Lattice2DItem*> latticeType() const;
    void setLatticeType(Lattice2DItem* p);

    bool xiIntegration() const;
    void setXiIntegration(bool xiIntegration);

protected:
    explicit Interference2DAbstractLatticeItem(bool xiIntegration);

    bool m_xiIntegration;
    SelectionProperty<Lattice2DItem*> m_latticeType;
};

// ------------------------------------------------------------------------------------------------

class Interference2DLatticeItem : public Interference2DAbstractLatticeItem {
public:
    Interference2DLatticeItem();
    std::unique_ptr<IInterference> createInterference() const override;
    void serialize(Streamer& s) override;

    void setDecayFunctionType(Profile2DItem* p);
    SelectionDescriptor<Profile2DItem*> decayFunction() const;

protected:
    SelectionProperty<Profile2DItem*> m_decayFunction;
};

// ------------------------------------------------------------------------------------------------

class Interference2DParaCrystalItem : public Interference2DAbstractLatticeItem {
public:
    Interference2DParaCrystalItem();
    std::unique_ptr<IInterference> createInterference() const override;
    void serialize(Streamer& s) override;

    DoubleDescriptor dampingLength() const;
    void setDampingLength(double dampingLength);

    DoubleDescriptor domainSize1() const;
    void setDomainSize1(double size);
    DoubleDescriptor domainSize2() const;
    void setDomainSize2(double size);

    SelectionDescriptor<Profile2DItem*> probabilityDistribution1() const;
    void setPDF1Type(Profile2DItem* p);

    SelectionDescriptor<Profile2DItem*> probabilityDistribution2() const;
    void setPDF2Type(Profile2DItem* p);

private:
    DoubleProperty m_dampingLength;
    DoubleProperty m_domainSize1;
    DoubleProperty m_domainSize2;
    SelectionProperty<Profile2DItem*> m_pdf1;
    SelectionProperty<Profile2DItem*> m_pdf2;
};

// ------------------------------------------------------------------------------------------------

class InterferenceFinite2DLatticeItem : public Interference2DAbstractLatticeItem {
public:
    InterferenceFinite2DLatticeItem();
    std::unique_ptr<IInterference> createInterference() const override;
    void serialize(Streamer& s) override;

    UIntDescriptor domainSize1() const;
    void setDomainSize1(unsigned int domain_size1);
    UIntDescriptor domainSize2() const;
    void setDomainSize2(unsigned int domain_size2);

private:
    UIntProperty m_domainSize1;
    UIntProperty m_domainSize2;
};

// ------------------------------------------------------------------------------------------------

class InterferenceHardDiskItem : public InterferenceItem {
public:
    InterferenceHardDiskItem();
    std::unique_ptr<IInterference> createInterference() const override;
    void serialize(Streamer& s) override;

    DoubleDescriptor radius() const;
    DoubleDescriptor density() const;

private:
    DoubleProperty m_radius;
    DoubleProperty m_density;
};

// ------------------------------------------------------------------------------------------------

class InterferenceRadialParaCrystalItem : public InterferenceItem {
public:
    InterferenceRadialParaCrystalItem();
    std::unique_ptr<IInterference> createInterference() const override;
    void serialize(Streamer& s) override;

    DoubleDescriptor peakDistance() const;
    DoubleDescriptor dampingLength() const;
    DoubleDescriptor domainSize() const;
    DoubleDescriptor kappa() const;
    SelectionDescriptor<Profile1DItem*> probabilityDistribution() const;
    void setPDFType(Profile1DItem* p);

private:
    DoubleProperty m_peakDistance;
    DoubleProperty m_dampingLength;
    DoubleProperty m_domainSize;
    DoubleProperty m_kappa;
    SelectionProperty<Profile1DItem*> m_pdf;
};

#endif // BORNAGAIN_GUI_MODEL_SAMPLE_INTERFERENCEITEMS_H
