//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Import/RealDataPropertiesWidget.cpp
//! @brief     Implements class RealDataPropertiesWidget
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2021
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Import/RealDataPropertiesWidget.h"
#include "GUI/Model/Data/RealDataItem.h"
#include "GUI/Model/Device/InstrumentItems.h"
#include "GUI/Model/Device/InstrumentsEditController.h"
#include "GUI/Model/Project/LinkInstrumentManager.h"
#include "GUI/Model/Project/ProjectDocument.h"
#include "GUI/View/Global/Globals.h"
#include <QComboBox>
#include <QLabel>
#include <QVBoxLayout>

RealDataPropertiesWidget::RealDataPropertiesWidget(QWidget* parent, ProjectDocument* document)
    : QWidget(parent)
    , m_instrumentCombo(new QComboBox)
    , m_currentDataItem(nullptr)
    , m_document(document)
{
    setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Preferred);
    setWindowTitle("RealDataPropertiesWidget");

    auto* mainLayout = new QVBoxLayout;
    mainLayout->setMargin(5);
    mainLayout->setSpacing(2);

    m_instrumentCombo->setToolTip("Select instrument to link with real data");

    mainLayout->addWidget(new QLabel("Linked instrument"));
    mainLayout->addWidget(m_instrumentCombo);

    mainLayout->addStretch();
    setLayout(mainLayout);

    updateInstrumentComboEntries();
    setEnabled(false);

    connect(m_instrumentCombo,
            static_cast<void (QComboBox::*)(int)>(&QComboBox::currentIndexChanged), this,
            &RealDataPropertiesWidget::onInstrumentComboIndexChanged);

    connect(m_document->instrumentsEditController(),
            &InstrumentsEditController::instrumentAddedOrRemoved, this,
            &RealDataPropertiesWidget::updateInstrumentComboEntries);

    connect(m_document->instrumentsEditController(),
            &InstrumentsEditController::instrumentNameChanged, this,
            &RealDataPropertiesWidget::updateInstrumentComboEntries);

    connect(m_document->linkInstrumentManager(), &LinkInstrumentManager::linkToInstrumentChanged,
            this, &RealDataPropertiesWidget::updateInstrumentComboIndex);
}

void RealDataPropertiesWidget::setItem(RealDataItem* item)
{
    m_currentDataItem = item;
    updateInstrumentComboIndex();
    setEnabled(m_currentDataItem != nullptr);
}

void RealDataPropertiesWidget::onInstrumentComboIndexChanged(int /*index*/)
{
    if (!m_currentDataItem)
        return;

    const QString newSelectedInstrumentId = m_instrumentCombo->currentData().toString();

    if (newSelectedInstrumentId == m_currentDataItem->instrumentId())
        return;

    if (m_document->linkInstrumentManager()->canLinkDataToInstrument(
            m_currentDataItem, newSelectedInstrumentId, GUI::Global::mainWindow)) {
        const auto* newSelectedInstrument =
            m_document->collectedItems()->findInstrumentById(newSelectedInstrumentId);
        m_currentDataItem->linkToInstrument(newSelectedInstrument);
    } else
        // Linking was impossible or denied. Set combo to previous state
        updateInstrumentComboIndex();
}

//! Updates instrument selector for new instruments and their names.
//! Current selection will be preserved.

void RealDataPropertiesWidget::updateInstrumentComboEntries()
{
    QSignalBlocker b(m_instrumentCombo);

    m_instrumentCombo->clear();

    // fill the combo. Userdata contains instrument's uid
    m_instrumentCombo->addItem("Undefined", ""); // undefined instrument
    for (auto* instrumentItem : m_document->collectedItems()->collectedItems())
        m_instrumentCombo->addItem(instrumentItem->instrumentName(), instrumentItem->id());

    updateInstrumentComboIndex();
}

//! Sets instrument combo selector to the state corresponding to given instrument identifier.

void RealDataPropertiesWidget::updateInstrumentComboIndex()
{
    const QString instrumentId =
        m_currentDataItem != nullptr ? m_currentDataItem->instrumentId() : "";

    QSignalBlocker b(m_instrumentCombo);
    m_instrumentCombo->setCurrentIndex(m_instrumentCombo->findData(instrumentId));
}
