//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Instrument/RectangularDetectorEditor.cpp
//! @brief     Implements class RectangularDetectorEditor
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Instrument/RectangularDetectorEditor.h"
#include "GUI/Model/Device/AxesItems.h"
#include "GUI/Model/Device/RectangularDetectorItem.h"
#include "GUI/View/Common/DoubleSpinBox.h"
#include "GUI/View/Instrument/DetectorAlignmentEditor.h"
#include "GUI/View/Instrument/ResolutionFunctionEditor.h"
#include "GUI/View/Tool/GroupBoxCollapser.h"
#include <QFormLayout>
#include <QGridLayout>
#include <QGroupBox>
#include <QSpinBox>
#include <QVBoxLayout>

RectangularDetectorEditor::RectangularDetectorEditor(QWidget* parent,
                                                     RectangularDetectorItem* detector)
    : QWidget(parent)
{
    ASSERT(detector);

    auto* grid = new QGridLayout(this);
    grid->setColumnStretch(0, 1);
    grid->setColumnStretch(1, 1);
    grid->setColumnStretch(2, 1);

    // -- x-axis controls
    auto* xAxisGroupBox = new QGroupBox("X axis", this);
    xAxisGroupBox->setProperty("subgroup", true); // for stylesheet addressing
    auto* xAxisFormLayout = new QFormLayout(xAxisGroupBox);

    auto* xAxisNbinsSpinBox = new QSpinBox(xAxisGroupBox);
    xAxisNbinsSpinBox->setRange(1, 65536);
    xAxisNbinsSpinBox->setValue(detector->xSize());
    xAxisFormLayout->addRow("Nbins:", xAxisNbinsSpinBox);

    auto* widthSpinBox = new DoubleSpinBox(xAxisGroupBox, detector->width());
    xAxisFormLayout->addRow("Width [mm]:", widthSpinBox);

    connect(xAxisNbinsSpinBox, qOverload<int>(&QSpinBox::valueChanged), [=](int newValue) {
        detector->setXSize(newValue);
        emit dataChanged();
    });

    connect(widthSpinBox, qOverload<double>(&DoubleSpinBox::baseValueChanged),
            [=](double newValue) {
                detector->setWidth(newValue);
                emit dataChanged();
            });
    grid->addWidget(xAxisGroupBox, 1, 0);
    GroupBoxCollapser::installIntoGroupBox(xAxisGroupBox);


    // -- y-axis controls
    auto* yAxisGroupBox = new QGroupBox("Y axis", this);
    yAxisGroupBox->setProperty("subgroup", true); // for stylesheet addressing
    auto* yAxisFormLayout = new QFormLayout(yAxisGroupBox);

    auto* yAxisNbinsSpinBox = new QSpinBox(yAxisGroupBox);
    yAxisNbinsSpinBox->setRange(1, 65536);
    yAxisNbinsSpinBox->setValue(detector->ySize());
    yAxisFormLayout->addRow("Nbins:", yAxisNbinsSpinBox);

    auto* heightSpinBox = new DoubleSpinBox(yAxisGroupBox, detector->height());
    yAxisFormLayout->addRow("Height [mm]:", heightSpinBox);

    connect(yAxisNbinsSpinBox, qOverload<int>(&QSpinBox::valueChanged), [=](int newValue) {
        detector->setYSize(newValue);
        emit dataChanged();
    });

    connect(heightSpinBox, qOverload<double>(&DoubleSpinBox::baseValueChanged),
            [=](double newValue) {
                detector->setHeight(newValue);
                emit dataChanged();
            });
    grid->addWidget(yAxisGroupBox, 1, 1);
    GroupBoxCollapser::installIntoGroupBox(yAxisGroupBox);

    // -- resolution function
    auto* resolutionFunctionEditor =
        new ResolutionFunctionEditor(ResolutionFunctionEditor::Millimeter, this, detector);
    grid->addWidget(resolutionFunctionEditor, 1, 2);

    connect(resolutionFunctionEditor, &ResolutionFunctionEditor::dataChanged, this,
            &RectangularDetectorEditor::dataChanged);

    // alignment selector editors
    auto* alignmentEditor = new DetectorAlignmentEditor(this, detector);
    grid->addWidget(alignmentEditor, 2, 0, 1, 3);

    connect(alignmentEditor, &DetectorAlignmentEditor::dataChanged, this,
            &RectangularDetectorEditor::dataChanged);
}
