//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Instrument/SphericalAxisEditor.cpp
//! @brief     Implement class SphericalAxisEditor
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Instrument/SphericalAxisEditor.h"
#include "GUI/Model/Device/AxesItems.h"
#include <QDoubleSpinBox>
#include <QFormLayout>
#include <QGroupBox>
#include <QSpinBox>

SphericalAxisForm::SphericalAxisForm(QFormLayout* form, QWidget* parent)
    : QObject(parent)
    , m_item(nullptr)
{
    m_nbinsSpinBox = new QSpinBox(parent);
    m_nbinsSpinBox->setRange(1, 65536);
    connect(m_nbinsSpinBox, qOverload<int>(&QSpinBox::valueChanged), this,
            &SphericalAxisForm::onNbinsValueChanged);
    form->addRow("Nbins:", m_nbinsSpinBox);

    m_minimumSpinBox = new QDoubleSpinBox(parent);
    m_minimumSpinBox->setRange(0, 90);
    m_minimumSpinBox->setDecimals(3);
    m_minimumSpinBox->setSingleStep(0.01);
    connect(m_minimumSpinBox, qOverload<double>(&QDoubleSpinBox::valueChanged), this,
            &SphericalAxisForm::onMinimumValueChanged);
    form->addRow("Minimum [deg]:", m_minimumSpinBox);

    m_maximumSpinBox = new QDoubleSpinBox(parent);
    m_maximumSpinBox->setRange(0, 90);
    m_maximumSpinBox->setDecimals(3);
    m_maximumSpinBox->setSingleStep(0.01);
    connect(m_maximumSpinBox, qOverload<double>(&QDoubleSpinBox::valueChanged), this,
            &SphericalAxisForm::onMaximumValueChanged);
    form->addRow("Maximum [deg]:", m_maximumSpinBox);
}

void SphericalAxisForm::setItem(BasicAxisItem* item)
{
    m_item = item;
    setEnabled(m_item);
    updateData();
}

void SphericalAxisForm::setEnabled(bool enabled)
{
    m_nbinsSpinBox->setEnabled(enabled);
    m_minimumSpinBox->setEnabled(enabled);
    m_maximumSpinBox->setEnabled(enabled);
}

void SphericalAxisForm::updateData()
{
    QSignalBlocker b1(m_nbinsSpinBox);
    QSignalBlocker b2(m_minimumSpinBox);
    QSignalBlocker b3(m_maximumSpinBox);
    if (m_item) {
        m_nbinsSpinBox->setValue(m_item->binCount());
        m_minimumSpinBox->setValue(m_item->min());
        m_maximumSpinBox->setValue(m_item->max());
    } else {
        m_nbinsSpinBox->setValue(1);
        m_minimumSpinBox->setValue(0.0);
        m_maximumSpinBox->setValue(0.0);
    }
}

void SphericalAxisForm::onNbinsValueChanged(int value)
{
    if (m_item && m_item->binCount() != value) {
        m_item->setBinCount(value);
        emit dataChanged();
    }
}

void SphericalAxisForm::onMinimumValueChanged(double value)
{
    if (m_item && m_item->min() != value) {
        m_item->setLowerBound(value);
        emit dataChanged();
        if (m_item->max() < value)
            m_maximumSpinBox->setValue(value);
    }
}

void SphericalAxisForm::onMaximumValueChanged(double value)
{
    if (m_item && m_item->max() != value) {
        m_item->setUpperBound(value);
        emit dataChanged();
        if (m_item->min() > value)
            m_minimumSpinBox->setValue(value);
    }
}
