//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Job/JobPropertiesWidget.cpp
//! @brief     Implements class JobPropertiesWidget
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Job/JobPropertiesWidget.h"
#include "GUI/Model/Job/JobItem.h"
#include "GUI/View/Job/JobPropertiesTableModel.h"
#include "GUI/View/Tool/StyleUtils.h"
#include "GUI/View/Tool/mainwindow_constants.h"
#include <QTabBar>
#include <QTabWidget>
#include <QTextEdit>
#include <QTreeView>
#include <QVBoxLayout>

namespace {

enum ETabId { JOB_PROPERTIES, JOB_MESSAGES };
}

//==================================================================================================
// JobPropertiesWidget
//==================================================================================================

//--------------------------------------------------------------------------------------------------
// public member functions
//--------------------------------------------------------------------------------------------------

JobPropertiesWidget::JobPropertiesWidget(QWidget* parent, Qt::WindowFlags f)
    : QWidget(parent, f)
    , m_item(nullptr)
{
    setSizePolicy(QSizePolicy::Expanding, QSizePolicy::MinimumExpanding);
    setWindowTitle(GUI::Constants::JobPropertiesWidgetName);

    auto* layout = new QVBoxLayout(this);
    layout->setMargin(0);
    layout->setSpacing(0);
    layout->setContentsMargins(0, 0, 0, 0);

    m_tabWidget = new QTabWidget(this);
    m_tabWidget->setTabPosition(QTabWidget::South);
    layout->addWidget(m_tabWidget);

    m_propertiesView = new QTreeView(this);
    m_propertiesView->setRootIsDecorated(false);
    m_propertiesView->setAlternatingRowColors(true);
    m_tabWidget->insertTab(JOB_PROPERTIES, m_propertiesView, "Job Properties");
    m_propertiesModel = new JobPropertiesTableModel(this);
    m_propertiesView->setModel(m_propertiesModel);

    m_commentsEditor = new QTextEdit(this);
    m_tabWidget->insertTab(JOB_MESSAGES, m_commentsEditor, "Messages");

    connect(m_commentsEditor, &QTextEdit::textChanged, this,
            &JobPropertiesWidget::onCommentsEdited);
}

JobPropertiesWidget::~JobPropertiesWidget()
{
    if (m_item)
        disconnect(m_item, nullptr, this, nullptr);
}

QSize JobPropertiesWidget::sizeHint() const
{
    return QSize(GUI::Util::Style::PropertyPanelWidth(), GUI::Util::Style::PropertyPanelWidth());
}

QSize JobPropertiesWidget::minimumSizeHint() const
{
    return QSize(GUI::Util::Style::PropertyPanelWidth(), GUI::Util::Style::PropertyPanelWidth());
}

void JobPropertiesWidget::setItem(JobItem* item)
{
    if (m_item)
        disconnect(m_item, nullptr, this, nullptr);
    m_item = item;
    m_propertiesModel->setItem(m_item);
    if (m_item) {
        if (m_item->isFailed())
            m_tabWidget->tabBar()->setTabTextColor(JOB_MESSAGES, Qt::red);
        else
            m_tabWidget->tabBar()->setTabTextColor(JOB_MESSAGES, Qt::black);
        m_commentsEditor->setText(m_item->getComments());

        notifyJobPropertyChange();
    } else
        m_commentsEditor->clear();
}

//--------------------------------------------------------------------------------------------------
// private member functions
//--------------------------------------------------------------------------------------------------

void JobPropertiesWidget::notifyJobPropertyChange()
{
    connect(
        m_item, &JobItem::jobCommentsChanged, this,
        [=](const QString&) {
            if (m_item && m_item->getComments() != m_commentsEditor->toPlainText()) {
                m_commentsEditor->blockSignals(true);
                m_commentsEditor->setPlainText(m_item->getComments());
                m_commentsEditor->blockSignals(false);
            }
        },
        Qt::UniqueConnection);
}

//--------------------------------------------------------------------------------------------------
// private slots
//--------------------------------------------------------------------------------------------------

void JobPropertiesWidget::onCommentsEdited()
{
    if (m_item) {
        m_item->blockSignals(true);
        m_item->setComments(m_commentsEditor->toPlainText());
        m_item->blockSignals(false);
    }
}
