//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Resample/Processed/ParticleRegions.cpp
//! @brief     Implements function SampleUtils::Preprocessor::particleRegions
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Resample/Processed/ParticleRegions.h"
#include "Resample/Processed/Slicer.h"
#include "Resample/Slice/ZLimits.h"
#include "Sample/Aggregate/ParticleLayout.h"
#include "Sample/Multilayer/Layer.h"
#include "Sample/Multilayer/MultiLayer.h"
#include "Sample/Particle/IParticle.h"
#include <algorithm>
#include <cstddef>
#include <stdexcept>
#include <utility>

//  ************************************************************************************************
//  define local class
//  ************************************************************************************************

//! Helper class for the graded layer approximation. Generates limits for each layer, indicating
//! the region of the layer (along z) that contains particle(s)
//!
//! The constructor takes the bottom layer z-coordinates as parameter. This means that for N
//! layers, only N-1 coordinates need to be passed (the last layer is assumed to be semi-infinite).
//! @ingroup algorithms_internal

class LayerFillLimits {
public:
    LayerFillLimits(std::vector<double> layers_bottomz);

    //! Particle limits are given in global coordinates
    void update(ZLimits particle_limits, double offset = 0.0);

    //! Returns the filled region limits for each layer (in local layer coordinates)
    std::vector<ZLimits> layerZLimits() const;

private:
    size_t i_layerTop(double top_z) const;
    size_t i_layerBottom(double bottom_z) const;
    void updateLayerLimits(size_t i_layer, ZLimits limits);
    std::vector<double> m_layers_bottomz;
    std::vector<ZLimits> m_layer_fill_limits;
};

#ifndef INCLUDED_BY_TEST

//  ************************************************************************************************
//  implement local class
//  ************************************************************************************************

namespace {

std::vector<double> bottomLayerCoordinates(const MultiLayer& sample)
{
    const size_t n_layers = sample.numberOfLayers();
    if (n_layers < 2)
        return {};
    std::vector<double> result(n_layers - 1);
    result[0] = 0.0;
    for (size_t i = 1; i < n_layers - 1; ++i)
        result[i] = result[i - 1] - sample.layer(i)->thickness();
    return result;
}

ZLimits CalculateNewLayerLimits(ZLimits old_limits, ZLimits bounded_limits)
{
    if (!old_limits.isFinite())
        return bounded_limits;
    return ZLimits::enclosingInterval(old_limits, bounded_limits);
}

} // namespace


LayerFillLimits::LayerFillLimits(std::vector<double> layers_bottomz)
    : m_layers_bottomz(std::move(layers_bottomz))
    , m_layer_fill_limits(m_layers_bottomz.size() + 1)
// default ZLimits designate an absence of limits
{
}

void LayerFillLimits::update(ZLimits particle_limits, double offset)
{
    if (m_layers_bottomz.empty())
        return; // do nothing for the single layer case
    double top = particle_limits.zTop() + offset;
    double bottom = particle_limits.zBottom() + offset;
    if (bottom > top)
        throw std::runtime_error("LayerFillLimits::update: lower_limit > upper_limit.");
    if (bottom == top) // zero-size particle
        return;
    size_t top_index = i_layerTop(top);
    size_t bottom_index = i_layerBottom(bottom);
    for (size_t i_layer = top_index; i_layer < bottom_index + 1; ++i_layer) {
        ZLimits limits(bottom, top);
        updateLayerLimits(i_layer, limits);
    }
}

std::vector<ZLimits> LayerFillLimits::layerZLimits() const
{
    return m_layer_fill_limits;
}

size_t LayerFillLimits::i_layerTop(double top_z) const
{
    if (m_layers_bottomz.empty())
        return 0;
    if (top_z <= m_layers_bottomz.back())
        return m_layers_bottomz.size();
    auto index_above = std::lower_bound(m_layers_bottomz.rbegin(), m_layers_bottomz.rend(), top_z);
    return static_cast<size_t>(m_layers_bottomz.rend() - index_above);
}

size_t LayerFillLimits::i_layerBottom(double bottom_z) const
{
    if (m_layers_bottomz.empty())
        return 0;
    if (bottom_z < m_layers_bottomz.back())
        return m_layers_bottomz.size();
    auto index_below =
        std::upper_bound(m_layers_bottomz.rbegin(), m_layers_bottomz.rend(), bottom_z);
    return static_cast<size_t>(m_layers_bottomz.rend() - index_below);
}

void LayerFillLimits::updateLayerLimits(size_t i_layer, ZLimits limits)
{
    if (!limits.isFinite())
        throw std::runtime_error("LayerFillLimits::updateLayerLimits: given limits are not "
                                 "finite.");
    auto old_limits = m_layer_fill_limits[i_layer];
    double layer_ref = i_layer ? m_layers_bottomz[i_layer - 1] : m_layers_bottomz[i_layer];
    double upper = i_layer ? std::min(limits.zTop(), layer_ref) : limits.zTop();
    double lower = (i_layer == m_layer_fill_limits.size() - 1)
                       ? limits.zBottom()
                       : std::max(limits.zBottom(), m_layers_bottomz[i_layer]);
    ZLimits bounded_limits(lower - layer_ref, upper - layer_ref);
    m_layer_fill_limits[i_layer] = CalculateNewLayerLimits(old_limits, bounded_limits);
}

//  ************************************************************************************************
//  exported function
//  ************************************************************************************************

//! Calculate z-admixtures occupied by particles
std::vector<ZLimits> Compute::Slicing::particleRegions(const MultiLayer& sample, bool use_slicing)
{
    const std::vector<double> bottom_coords = bottomLayerCoordinates(sample);
    LayerFillLimits layer_fill_limits(bottom_coords);
    if (use_slicing) {
        for (size_t i = 0; i < sample.numberOfLayers(); ++i) {
            const Layer* layer = sample.layer(i);
            const double offset = (i == 0) ? 0 : bottom_coords[i - 1];
            for (const auto* layout : layer->layouts())
                for (const IParticle* particle : layout->particles())
                    layer_fill_limits.update(Compute::Slicing::zSpan(particle), offset);
        }
    }
    return layer_fill_limits.layerZLimits();
}

#endif // INCLUDED_BY_TEST
