//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Sample/Aggregate/ParticleLayout.h
//! @brief     Defines class ParticleLayout.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_SAMPLE_AGGREGATE_PARTICLELAYOUT_H
#define BORNAGAIN_SAMPLE_AGGREGATE_PARTICLELAYOUT_H

#include "Base/Types/OwningVector.h"
#include "Sample/Scattering/ISampleNode.h"
#include <memory>

class IInterference;
class IParticle;

//! Decorator class that adds particles to ISampleNode objects.
//! @ingroup samples

class ParticleLayout : public ISampleNode {
public:
    ParticleLayout();
    ParticleLayout(const IParticle& particle, double abundance = -1.0);
    ~ParticleLayout() override;

    ParticleLayout* clone() const override;
    std::string className() const final { return "ParticleLayout"; }
    std::vector<ParaMeta> parDefs() const final
    {
        return {{"Abundance", "", "e.g. surface fraction covered by this layout", 0, +INF, -1}};
    }
    std::vector<const INode*> nodeChildren() const override;

    void addParticle(const IParticle& particle, double abundance = -1.0);

    std::vector<const IParticle*> particles() const;

    const IInterference* interferenceFunction() const;

    double totalAbundance() const;

    void setInterference(const IInterference& interparticle);

    double totalParticleSurfaceDensity() const;
    void setTotalParticleSurfaceDensity(double particle_density);

    //! Returns the relative weight of this layout
    double weight() const { return m_weight; }

    //! Sets the relative weight of this layout
    void setWeight(double weight) { m_weight = weight; }

    double weightedParticleSurfaceDensity() const;

private:
    void addAndRegisterAbstractParticle(IParticle* child);
    void setAndRegisterInterference(IInterference* child);

    double m_weight{1};
    double m_total_particle_density{.01};
    OwningVector<IParticle> m_particles; //!< Vector of particle types
    std::unique_ptr<IInterference> m_interparticle;
};

#endif // BORNAGAIN_SAMPLE_AGGREGATE_PARTICLELAYOUT_H
