//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Sample/Correlations/Profiles1D.h
//! @brief     Defines interface class IProfile1D, and children thereof
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_SAMPLE_CORRELATIONS_PROFILES1D_H
#define BORNAGAIN_SAMPLE_CORRELATIONS_PROFILES1D_H

#include "Base/Types/ICloneable.h"
#include "Param/Node/INode.h"
#ifndef SWIG
#include "Sample/Correlations/IDistribution1DSampler.h"
#endif // SWIG

#ifndef USER_API

//! Interface for a one-dimensional distribution, with normalization adjusted so that
//! the Fourier transform standardizedFT(q) is a decay function that starts at standardizedFT(0)=1.

class IProfile1D : public ICloneable, public INode {
public:
    IProfile1D(const std::vector<double>& PValues);

    IProfile1D* clone() const override = 0;

    //! Returns Fourier transform of the normalized distribution;
    //! is a decay function starting at standardizedFT(0)=1.
    virtual double standardizedFT(double q) const = 0;
    //! Returns Fourier transform of the distribution scaled as decay function f(x)/f(0).
    virtual double decayFT(double q) const = 0;

    double omega() const { return m_omega; }
    double decayLength() const { return m_omega; }

    //! Returns the negative of the second order derivative in q space around q=0
    virtual double qSecondDerivative() const = 0;
#ifndef SWIG
    virtual std::unique_ptr<IDistribution1DSampler> createSampler() const = 0;

    //! Creates the Python constructor of this class (or derived classes)
    virtual std::string pythonConstructor() const;
#endif

protected:
    const double& m_omega;
};

#endif // USER_API

//! Exponential IProfile1D exp(-|omega*x|);
//! its Fourier transform standardizedFT(q) is a Cauchy-Lorentzian starting at standardizedFT(0)=1.
//! @ingroup distributionFT

class Profile1DCauchy : public IProfile1D {
public:
    Profile1DCauchy(std::vector<double> P);
    Profile1DCauchy(double omega);

    Profile1DCauchy* clone() const override;
    std::string className() const final { return "Profile1DCauchy"; }
    // const auto tooltip = "class_tooltip";
    std::vector<ParaMeta> parDefs() const final
    {
        return {{"Omega", "nm", "Half-width", 0, INF, 1.}};
    }
    double standardizedFT(double q) const override;
    double decayFT(double q) const override;
    double qSecondDerivative() const override;
#ifndef SWIG
    std::unique_ptr<IDistribution1DSampler> createSampler() const override;
#endif
};

//! Gaussian IProfile1D;
//! its Fourier transform standardizedFT(q) is a Gaussian starting at standardizedFT(0)=1.
//! @ingroup distributionFT

class Profile1DGauss : public IProfile1D {
public:
    Profile1DGauss(std::vector<double> P);
    Profile1DGauss(double omega);

    Profile1DGauss* clone() const override;
    std::string className() const final { return "Profile1DGauss"; }
    // const auto tooltip = "class_tooltip";
    std::vector<ParaMeta> parDefs() const final
    {
        return {{"Omega", "nm", "Half-width", 0, INF, 1.}};
    }
    double standardizedFT(double q) const override;
    double decayFT(double q) const override;
    double qSecondDerivative() const override;
#ifndef SWIG
    std::unique_ptr<IDistribution1DSampler> createSampler() const override;
#endif
};

//! Square gate IProfile1D;
//! its Fourier transform standardizedFT(q) is a sinc function starting at standardizedFT(0)=1.
//! @ingroup distributionFT

class Profile1DGate : public IProfile1D {
public:
    Profile1DGate(std::vector<double> P);
    Profile1DGate(double omega);

    Profile1DGate* clone() const override;
    std::string className() const final { return "Profile1DGate"; }
    // const auto tooltip = "class_tooltip";
    std::vector<ParaMeta> parDefs() const final
    {
        return {{"Omega", "nm", "Half-width", 0, INF, 1.}};
    }
    double standardizedFT(double q) const override;
    double decayFT(double q) const override;
    double qSecondDerivative() const override;
#ifndef SWIG
    std::unique_ptr<IDistribution1DSampler> createSampler() const override;
#endif
};

//! Triangle IProfile1D [1-|x|/omega if |x|<omega, and 0 otherwise];
//! its Fourier transform standardizedFT(q) is a squared sinc function starting at
//! standardizedFT(0)=1.
//! @ingroup distributionFT

class Profile1DTriangle : public IProfile1D {
public:
    Profile1DTriangle(std::vector<double> P);
    Profile1DTriangle(double omega);

    Profile1DTriangle* clone() const override;
    std::string className() const final { return "Profile1DTriangle"; }
    // const auto tooltip = "class_tooltip";
    std::vector<ParaMeta> parDefs() const final
    {
        return {{"Omega", "nm", "Half-width", 0, INF, 1.}};
    }
    double standardizedFT(double q) const override;
    double decayFT(double q) const override;
    double qSecondDerivative() const override;
#ifndef SWIG
    std::unique_ptr<IDistribution1DSampler> createSampler() const override;
#endif
};

//! IProfile1D consisting of one cosine wave
//! [1+cos(pi*x/omega) if |x|<omega, and 0 otherwise];
//! its Fourier transform standardizedFT(q) starts at standardizedFT(0)=1.
//! @ingroup distributionFT

class Profile1DCosine : public IProfile1D {
public:
    Profile1DCosine(std::vector<double> P);
    Profile1DCosine(double omega);

    Profile1DCosine* clone() const override;
    std::string className() const final { return "Profile1DCosine"; }
    // const auto tooltip = "class_tooltip";
    std::vector<ParaMeta> parDefs() const final
    {
        return {{"Omega", "nm", "Half-width", 0, INF, 1.}};
    }
    double standardizedFT(double q) const override;
    double decayFT(double q) const override;
    double qSecondDerivative() const override;
#ifndef SWIG
    std::unique_ptr<IDistribution1DSampler> createSampler() const override;
#endif
};

//! IProfile1D that provides a Fourier transform standardizedFT(q) in form
//! of a pseudo-Voigt decay function eta*Gauss + (1-eta)*Cauchy, with both components
//! starting at 1 for q=0.
//! @ingroup distributionFT

class Profile1DVoigt : public IProfile1D {
public:
    Profile1DVoigt(std::vector<double> P);
    Profile1DVoigt(double omega, double eta);

    Profile1DVoigt* clone() const override;
    std::string className() const final { return "Profile1DVoigt"; }
    // const auto tooltip = "Pseudo-Voigt distribution";
    std::vector<ParaMeta> parDefs() const final
    {
        return {{"Omega", "nm", "Half-width", 0, INF, 1.},
                {"Eta", "", "balances between Gauss (eta=0) and Lorentz (eta=1) limiting cases", 0,
                 1, .5}};
    }
    double standardizedFT(double q) const override;
    double decayFT(double q) const override;
    double eta() const { return m_eta; }
    double qSecondDerivative() const override;
#ifndef SWIG
    std::unique_ptr<IDistribution1DSampler> createSampler() const override;
    std::string pythonConstructor() const override;
#endif

protected:
    const double& m_eta;
};

#endif // BORNAGAIN_SAMPLE_CORRELATIONS_PROFILES1D_H
