//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Sample/HardParticle/Pyramid4.cpp
//! @brief     Implements class Pyramid4.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Sample/HardParticle/Pyramid4.h"
#include "Base/Math/Constants.h"
#include "Base/Math/Functions.h"

const ff::PolyhedralTopology Pyramid4::topology = {{
                                                       {{3, 2, 1, 0}, true}, // TODO -> true
                                                       {{0, 1, 5, 4}, false},
                                                       {{1, 2, 6, 5}, false},
                                                       {{2, 3, 7, 6}, false},
                                                       {{3, 0, 4, 7}, false},
                                                       {{4, 5, 6, 7}, true} // TODO -> true
                                                   },
                                                   false};

Pyramid4::Pyramid4(const std::vector<double> P)
    : IFormFactorPolyhedron(P)
    , m_base_edge(m_P[0])
    , m_height(m_P[1])
    , m_alpha(m_P[2])
{
    checkNodeArgs();
    double cot_alpha = Math::cot(m_alpha);
    if (!std::isfinite(cot_alpha))
        throw std::runtime_error("pyramid angle alpha out of bounds");
    double r = cot_alpha * 2 * m_height / m_base_edge; // [L(base)-L(top)]/L(base)
    if (r > 1) {
        std::ostringstream ostr;
        ostr << "Pyramid4() -> Error in class initialization with parameters";
        ostr << " base_edge:" << m_base_edge;
        ostr << " height:" << m_height;
        ostr << " alpha[rad]:" << m_alpha << "\n\n";
        ostr << "Check for 'height <= base_edge*tan(alpha)' failed.";
        throw std::runtime_error(ostr.str());
    }

    double a = m_base_edge / 2;
    double b = a * (1 - r);

    double zcom = m_height * (.5 - 2 * r / 3 + r * r / 4) / (1 - r + r * r / 3); // center of mass

    setPolyhedron(topology, -zcom,
                  {// base:
                   {-a, -a, -zcom},
                   {a, -a, -zcom},
                   {a, a, -zcom},
                   {-a, a, -zcom},
                   // top:
                   {-b, -b, m_height - zcom},
                   {b, -b, m_height - zcom},
                   {b, b, m_height - zcom},
                   {-b, b, m_height - zcom}});
}

Pyramid4::Pyramid4(double base_edge, double height, double alpha)
    : Pyramid4(std::vector<double>{base_edge, height, alpha})
{
}
